# coding: utf-8
import os
from collections import OrderedDict
import numpy as np
import pandas as pd
from tqdm import tqdm
from pyecharts.charts import Page
from typing import List, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
from ..utils.echarts_plot import box_plot

plt.style.use('ggplot')
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False


def prepare_factors(factors: pd.DataFrame, bar_numbers: [List, Tuple] = (1, 2, 5, 10, 20, 30, 60, 120, 240)):
    """预处理

    :param factors:
    :param bar_numbers:
    :return:
    """
    factors['base'] = 1
    price_column = 'close'
    for bar_number in bar_numbers:
        n_col_name = 'n' + str(bar_number) + 'b'
        factors[n_col_name] = (factors[price_column].shift(-bar_number) / factors[price_column] - 1) * 10000
    factors['dt'] = pd.to_datetime(factors['dt'])
    return factors


def expand_category(factors: pd.DataFrame, category_cols):
    for f in category_cols:
        if factors[f].dtype == 'O' or factors[f].dtype == 'bool':
            itm_list = factors[f].unique()
            for itm in itm_list:
                ecol = '%s@%s' % (f, itm)
                factors[ecol] = np.where(factors[f] == itm, 1, 0)
    return factors


def factors_to_bs(factors: pd.DataFrame, opens: List, exits: List, direction="long", cost=0.3):
    """从因子生成交易序列

    :param cost: 每一个交易对的固定交易成本
    :param factors:
    :param opens:
    :param exits:
    :param direction:
    :return:
    """
    name = opens[0].split("@")[0]
    last_op = "exit"
    ops = []
    for i, row in tqdm(factors.iterrows()):
        op = {"dt": row['dt'], "symbol": row['symbol'], "price": row['close'], "direction": direction}
        if last_op == 'exit' and f"{name}@{row[name]}" in opens:
            op.update({'op': "long_open", 'op_detail': f"{name}@{row[name]}"})
            last_op = 'open'
            ops.append(op)

        elif last_op == 'open' and f"{name}@{row[name]}" in exits:
            op.update({'op': "long_exit", 'op_detail': f"{name}@{row[name]}"})
            last_op = 'exit'
            ops.append(op)
    if ops:
        df = pd.DataFrame(ops)
    else:
        df = pd.DataFrame()

    # 构造成交易对形式输出
    if len(ops) % 2 != 0:
        ops.pop(-1)

    pairs = []
    for i in range(0, len(ops), 2):
        b, s = ops[i: i+2]
        pair = OrderedDict({
            "标的代码": b['symbol'],
            "开仓时间": b['dt'],
            "开仓价格": b['price'],
            "开仓因子": b['op_detail'],
            "平仓时间": s['dt'],
            "平仓价格": s['price'],
            "平仓因子": s['op_detail'],
            "持仓天数": (s['dt'] - b['dt']).days,
            "盈亏(%)": round((s['price'] - b['price']) / b['price'] * 100 - cost, 2),
        })

        if pairs:
            pair['净值(%)'] = round((pairs[-1]['净值(%)']) * (1 + pair['盈亏(%)'] / 100), 2)
        else:
            pair['净值(%)'] = round(100 * (1 + pair['盈亏(%)'] / 100), 2)
        pairs.append(pair)
    if pairs:
        df1 = pd.DataFrame(pairs)
    else:
        df1 = pd.DataFrame()

    return df, df1


def report_factor_performance(factors: pd.DataFrame, name: str):
    """评估单个因子的表现

    :param factors:
    :param name: 因子名称
    :return:
    """
    nb_cols = [x for x in factors.columns if x.startswith("n")]
    page = Page(layout=Page.DraggablePageLayout, page_title=f"{name} - 性能分析")
    box_data = {col: [factors[col].quantile(v) for v in [0, 0.25, 0.5, 0.75, 1]] for col in nb_cols}
    chart = box_plot(data=box_data, title=f"BASE", width=f'{40*len(nb_cols)}px', height='300px')
    page.add(chart)
    for k, df in factors.groupby(name):
        box_data = {col: [df[col].quantile(v) for v in [0, 0.25, 0.5, 0.75, 1]] for col in nb_cols}
        chart = box_plot(data=box_data, title=f"{name}@{k}", width=f'{40*len(nb_cols)}px', height='300px')
        page.add(chart)
    return page

class FactorsResearcher:
    """因子研究员"""

    def __init__(self, factors, n_list=None):
        if isinstance(factors, pd.DataFrame):
            self.factors = factors
        else:
            self.factors = pd.DataFrame(factors)
        self.factors['base'] = prepare_factors(self.factors, self.n_list)
        if not n_list:
            self.n_list = (15, 30, 60, 120, 240, 480, 960)
        else:
            self.n_list = n_list
        self.factors = prepare_factors(self.factors, self.n_list)
        self.base = self.factor_performance(['base'], self.n_list).iloc[0].to_dict()
        self.symbol = self.base['symbol']

    def factor_performance(self, cols: List, n_list: [List, Tuple]):
        """查看因子组合的表现"""
        df = self.factors
        df = df.dropna()
        n_list = list(n_list)

        results = []
        for values, dfg in df.groupby(cols):
            res = {"symbol": self.factors[0]['symbol'], "factor_name": "#".join(cols)}
            if isinstance(values, tuple):
                res['factor_value'] = "#".join(values)
            else:
                res['factor_value'] = values

            res["count"] = len(dfg)

            for n in n_list:
                nb_col = 'n{}b'.format(n)
                dn = dfg[nb_col]
                dn1 = dfg[dfg[nb_col] >= 0]
                # dn2 = dfg[dfg[nb_col] < 0]
                res['n{}b均收益'.format(n)] = round(dn.mean(), 2)
                # if dn2.empty:
                #     res['n{}b盈亏比'.format(n)] = round(dn1[nb_col].mean(), 2)
                # elif dn1.empty:
                #     res['n{}b盈亏比'.format(n)] = round(dn2[nb_col].mean(), 2)
                # else:
                #     res['n{}b盈亏比'.format(n)] = round(abs(dn1[nb_col].mean() / (dn2[nb_col].mean() + 1)), 2)
                res['n{}b胜率'.format(n)] = round(len(dn1) / len(dn), 2)

            results.append(res)
        df_res = pd.DataFrame(results)
        df_res['pct'] = df_res['count'] / len(df)
        df_res['pct'] = df_res['pct'].round(4)
        return df_res

    def plot_performance(self, cols: List, show: bool = True, res_path: str = None, dpi: int = 100):
        """绘制nbars表现图"""
        base = self.base
        n_bars = [x for x in base.keys() if "均收益" in x]
        base_v = [base[x] for x in n_bars]
        df = self.factor_performance(cols, self.n_list)
        plt.close()
        ncols = 4
        nrows = len(df) // ncols if len(df) % ncols == 0 else len(df) // ncols + 1
        fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=(6*ncols, len(df)*4 // ncols))
        for i, v in enumerate(df.factor_value.unique()):
            f = df[df['factor_value'] == v].iloc[0].to_dict()
            factor_ = [f[x] for x in n_bars]
            data = {"n_bars": [int(x.strip("nb均收益")) for x in n_bars], 'base': base_v, v: factor_}

            ax = axes[i // ncols][i % ncols]
            sns.lineplot(x="n_bars", y='base', data=data, ax=ax)
            sns.lineplot(x="n_bars", y=v, data=data, ax=ax)
            ax.legend(["base", 'factor'])
            names = f['factor_name'].split("#")
            values = [x.split("~")[0] for x in f['factor_value'].split("#")]
            v = ["{}({})".format(x1, x2) for x1, x2 in zip(names, values)]
            ax.set_title(v, loc='center')
            ax.text(x=10, y=5, s="数量：{}({}%)".format(f['count'], round(f['pct']*100, 2)), fontsize=14)

        if show:
            plt.show()
        else:
            file_png = os.path.join(res_path, "{}_{}.png".format(self.symbol, "&".join(cols)))
            plt.savefig(file_png, bbox_inches='tight', dpi=dpi)

    def single_factor_analyze(self, cols: [list, tuple], n_list: [List, Tuple]):
        """单因子分析"""
        data = []
        for col in tqdm(cols, desc="single_factor_analyze"):
            data.append(self.factor_performance([col], n_list))
        df = pd.concat(data)
        return df

    def pair_factor_analyze(self, pairs:  List[Tuple[str, str]], n_list: [List, Tuple]):
        """两因子分析"""
        data = []
        for pair in tqdm(pairs, desc="pair_factor_analyze"):
            data.append(self.factor_performance(list(pair), n_list))
        df = pd.concat(data)
        return df

    def triple_factor_analyze(self, triples: List[Tuple[str, str, str]], n_list: [List, Tuple]):
        """三因子分析"""
        data = []
        for triple in tqdm(triples, desc="triple_factor_analyze"):
            data.append(self.factor_performance(list(triple), n_list))
        df = pd.concat(data)
        return df



# coding: utf-8
import warnings
from datetime import datetime, timedelta
from typing import List
from ..objects import RawBar


def bar_end_time(dt: datetime, m=1):
    """获取 dt 对应的分钟周期结束时间

    :param dt: datetime
    :param m: int
        分钟周期，1 表示 1分钟，5 表示 5分钟 ...
    :return: datetime
    """
    dt = dt.replace(second=0, microsecond=0)
    dt_span = {
        60: ["01:00", "2:00", "3:00", '10:30', "11:30", "14:00", "15:00", "22:00", "23:00", "23:59"],
    }

    if m < 60:
        if (dt.hour == 15 and dt.minute == 0) or (dt.hour == 11 and dt.minute == 30):
            return dt

        delta_m = dt.minute % m
        if delta_m != 0:
            dt += timedelta(minutes=m - delta_m)
        else:
            dt += timedelta(minutes=m)
        return dt
    else:
        for v in dt_span[m]:
            hour, minute = v.split(":")
            edt = dt.replace(hour=int(hour), minute=int(minute))
            if dt <= edt:
                return edt
    return dt

class KlineGeneratorBase:
    """K线生成器，仿实盘"""

    def __init__(self, max_count: int = 5000, freqs: List[str] = None):
        """

        :param max_count: int
            最大K线数量
        :param freqs: list of str
            级别列表，默认值为 ['周线', '日线', '60分钟', '30分钟', '15分钟', '5分钟', '1分钟']
        """
        self.max_count = max_count
        if freqs is None:
            self.freqs = ['周线', '日线', '60分钟', '30分钟', '15分钟', '5分钟', '1分钟']
        else:
            self.freqs = freqs
        self.m1: List[RawBar] = []
        self.m5: List[RawBar] = []
        self.m15: List[RawBar] = []
        self.m30: List[RawBar] = []
        self.m60: List[RawBar] = []
        self.D: List[RawBar] = []
        self.W: List[RawBar] = []
        self.end_dt = None
        self.symbol = None

    def __update_end_dt(self):
        if self.m1:
            self.end_dt = self.m1[-1].dt
            self.symbol = self.m1[-1].symbol

    def init_kline(self, freq: str, kline: List[RawBar]):
        """输入K线进行初始化

        :param freq: str
        :param kline: list of dict
        :return:
        """
        freqs_map = {"1分钟": self.m1, "5分钟": self.m5, "15分钟": self.m15,
                     "30分钟": self.m30, "60分钟": self.m60, "日线": self.D, "周线": self.W}
        m = freqs_map[freq]
        m.extend(kline)
        self.__update_end_dt()

    def __repr__(self):
        return "<KlineGenerator for {}; latest_dt={}>".format(self.symbol, self.end_dt)

    # def __update_minutes(self):
    #     pass
    #
    # def __update_d(self):
    #     pass
    #
    # def __update_w(self):
    #     pass
    #
    # def update(self):
    #     pass

    def get_kline(self, freq: str, count: int = 1000) -> List[RawBar]:
        """获取单个级别的K线

        :param freq: str
            级别名称，可选值 1分钟；5分钟；15分钟；30分钟；60分钟；日线；周线
        :param count: int
            数量
        :return: list of dict
        """
        freqs_map = {"1分钟": self.m1, "5分钟": self.m5, "15分钟": self.m15,
                     "30分钟": self.m30, "60分钟": self.m60, "日线": self.D, "周线": self.W}
        return freqs_map[freq][-count:]

    def get_klines(self, counts=None):
        """获取多个级别的K线

        :param counts: dict
            默认值 {"1分钟": 1000, "5分钟": 1000, "30分钟": 1000, "日线": 100}
        :return: dict of list of dict
        """
        if counts is None:
            counts = {"1分钟": 1000, "5分钟": 1000, "30分钟": 1000, "日线": 100}
        return {k: self.get_kline(k, v) for k, v in counts.items()}


class KlineGeneratorByTick(KlineGeneratorBase):
    """K线生成器，仿实盘，从tick开始生成"""

    def __init__(self, max_count: int = 5000, freqs: List[str] = None):
        """

        :param max_count: int
            最大K线数量
        :param freqs: list of str
            级别列表，默认值为 ['周线', '日线', '60分钟', '30分钟', '15分钟', '5分钟', '1分钟']
        """
        super().__init__(max_count, freqs)
        warnings.warn("KlineGeneratorByTick 存在已知Bug，慎用")

    def __repr__(self):
        return "<KlineGeneratorByTick for {}; latest_dt={}>".format(self.symbol, self.end_dt)

    @staticmethod
    def __update_from_tick(last, tick):
        new = dict(last)
        new.update({
            'close': round(tick['price'], 2),
            "dt": tick['dt'],
            "high": round(max(last['high'], tick['price']), 2),
            "low": round(min(last['low'], tick['price']), 2),
            "vol": last['vol'] + tick['vol']
        })
        if new['open'] <= 0:
            new['open'] = round(tick['price'], 2)

        if new['low'] <= 0:
            new['low'] = round(tick['price'], 2)

        return new

    @staticmethod
    def __init_bar_from_tick(tick):
        p = round(tick['price'], 2)
        return {
            'symbol': tick['symbol'],
            'open': p,
            'close': p,
            "dt": tick['dt'],
            "high": p,
            "low": p,
            "vol": tick['vol']
        }

    def __update_minutes(self, tick=None, minutes=(1, 5, 15, 30, 60)):
        # 更新分钟线
        fm_map = {1: self.m1, 5: self.m5, 15: self.m15, 30: self.m30, 60: self.m60}

        for minute in minutes:
            if "{}分钟".format(minute) not in self.freqs:
                continue

            m = fm_map[minute]
            if not m:
                next_bar = self.__init_bar_from_tick(tick)
                next_bar['dt'] = bar_end_time(tick['dt'], m=minute)
                m.append(next_bar)
            else:
                last = m[-1]
                next_end_dt = bar_end_time(tick['dt'], m=minute)
                if next_end_dt > last['dt'] and next_end_dt.minute != last['dt'].minute:
                    next_bar = self.__init_bar_from_tick(tick)
                    next_bar['dt'] = next_end_dt
                    m.append(next_bar)
                else:
                    next_bar = self.__update_from_tick(last, tick)
                    next_bar['dt'] = next_end_dt
                    m[-1] = next_bar
            fm_map[minute] = m[-self.max_count:]

    def __update_d(self, tick=None):
        if "日线" not in self.freqs:
            return

        if not self.D:
            self.D.append(self.__init_bar_from_tick(tick))
        else:
            last = self.D[-1]
            if last['dt'].date() != tick['dt'].date():
                self.D.append(self.__init_bar_from_tick(tick))
            else:
                self.D[-1] = self.__update_from_tick(last, tick)
        self.D = self.D[-self.max_count:]

    def __update_w(self, tick=None):
        if "周线" not in self.freqs:
            return

        if not self.W:
            self.W.append(self.__init_bar_from_tick(tick))
        else:
            last = self.W[-1]
            if tick['dt'].weekday() == 0 and tick['dt'].weekday() != last['dt'].weekday():
                self.W.append(self.__init_bar_from_tick(tick))
            else:
                self.W[-1] = self.__update_from_tick(last, tick)
        self.W = self.W[-self.max_count:]

    def update(self, tick=None):
        """输入1分钟最新K线 或 tick，更新其他级别K线

        :param tick: dict
            {'symbol': '000001.XSHG',
             'dt': Timestamp('2020-07-16 14:51:00'),
             'price': 3216.8,
             'vol': '270429600'}
        """
        if self.end_dt and tick['dt'] < self.end_dt:
            warnings.warn("输入 tick 时间小于最近一个更新时间，{} < {}，不进行K线更新".format(tick['dt'], self.end_dt))
            return

        assert isinstance(tick, dict)
        self.end_dt = tick['dt']
        self.symbol = tick['symbol']
        self.__update_minutes(tick, minutes=(1, 5, 15, 30, 60))
        if "日线" in self.freqs:
            self.__update_d(tick)
        if "周线" in self.freqs:
            self.__update_w(tick)


class KlineGeneratorBy1Min(KlineGeneratorBase):
    """K线生成器，仿实盘，从1分钟开始生成"""

    def __init__(self, max_count: int = 5000, freqs: List[str] = None):
        """

        :param max_count: int
            最大K线数量
        :param freqs: list of str
            级别列表，默认值为 ['周线', '日线', '60分钟', '30分钟', '15分钟', '5分钟', '1分钟']
        """
        super().__init__(max_count, freqs)

    def __repr__(self):
        return "<KlineGeneratorBy1Min for {}; latest_dt={}>".format(self.symbol, self.end_dt)

    @staticmethod
    def __update_from_1min(last: RawBar, k: RawBar, next_end_dt: datetime):
        new = RawBar(
            symbol=last.symbol,
            dt=next_end_dt,
            open=last.open,
            close=k.close,
            high=max(last.high, k.high),
            low=min(last.low, k.low),
            vol=last.vol + k.vol,
        )
        return new

    def __update_1min(self, k: RawBar):
        # 更新1分钟线
        if "1分钟" not in self.freqs:
            return

        if not self.m1:
            self.m1.append(k)
        else:
            if k.dt > self.m1[-1].dt:
                self.m1.append(k)
            elif k.dt == self.m1[-1].dt:
                self.m1[-1] = k
            else:
                raise ValueError("1分钟新K线的时间{}必须大于等于最后一根K线的时间{}".format(k.dt, self.m1[-1].dt))

        self.m1 = self.m1[-self.max_count:]

    def __update_minutes(self, k: RawBar, minutes=(5, 15, 30, 60)):
        # 更新分钟线
        fm_map = {5: self.m5, 15: self.m15, 30: self.m30, 60: self.m60}

        for minute in minutes:
            if "{}分钟".format(minute) not in self.freqs:
                continue

            m = fm_map[minute]
            next_end_dt = bar_end_time(k.dt, m=minute)
            if not m:
                m.append(RawBar(symbol=k.symbol, dt=next_end_dt, open=k.open, close=k.close,
                                high=k.high, low=k.low, vol=k.vol))
            else:
                last = m[-1]
                if next_end_dt != last.dt:
                    m.append(RawBar(symbol=k.symbol, dt=next_end_dt, open=k.open, close=k.close,
                                    high=k.high, low=k.low, vol=k.vol))
                else:
                    next_bar = self.__update_from_1min(last, k, next_end_dt)
                    m[-1] = next_bar
            fm_map[minute] = m[-self.max_count:]

    def __update_d(self, k=None):
        if "日线" not in self.freqs:
            return
        next_end_dt = k.dt.replace(hour=0, minute=0, second=0, microsecond=0)
        k = RawBar(symbol=k.symbol, dt=next_end_dt, open=k.open, close=k.close, high=k.high, low=k.low, vol=k.vol)
        if not self.D:
            self.D.append(k)
        last = self.D[-1]
        if next_end_dt.date() != last.dt.date():
            self.D.append(k)
        else:
            self.D[-1] = self.__update_from_1min(last, k, next_end_dt)

        self.D = self.D[-self.max_count:]

    def __update_w(self, k=None):
        if "周线" not in self.freqs:
            return
        next_end_dt = k.dt.replace(hour=0, minute=0, second=0, microsecond=0)
        k = RawBar(symbol=k.symbol, dt=next_end_dt, open=k.open, close=k.close, high=k.high, low=k.low, vol=k.vol)
        if not self.W:
            self.W.append(k)
        last = self.W[-1]
        if next_end_dt.weekday() == 0 and k.dt.weekday() != last.dt.weekday():
            self.W.append(k)
        else:
            self.W[-1] = self.__update_from_1min(last, k, next_end_dt)

        self.W = self.W[-self.max_count:]

    def update(self, k: RawBar):
        """输入1分钟最新K线，更新其他级别K线

        :param k: 1分钟K线
        """
        if self.end_dt and k.dt < self.end_dt:
            print("输入1分钟K时间小于最近一个更新时间，{} < {}，不进行K线更新".format(k.dt, self.end_dt))
            return

        self.end_dt = k.dt
        self.symbol = k.symbol

        self.__update_1min(k)
        self.__update_minutes(k, minutes=(5, 15, 30, 60))
        if "日线" in self.freqs:
            self.__update_d(k)
        if "周线" in self.freqs:
            self.__update_w(k)




# coding: utf-8
import sys
import warnings

sys.path.insert(0, '.')
sys.path.insert(0, '..')

import os
import pandas as pd
import random
from czsc.utils import echarts_plot as plot
from czsc.analyze import CZSC, RawBar, Mark

cur_path = os.path.split(os.path.realpath(__file__))[0]


def test_heat_map():
    data = [{"x": "{}hour".format(i), "y": "{}day".format(j), "heat": random.randint(0, 50)}
            for i in range(24) for j in range(7)]
    x_label = ["{}hour".format(i) for i in range(24)]
    y_label = ["{}day".format(i) for i in range(7)]
    hm = plot.heat_map(data, x_label=x_label, y_label=y_label)
    hm.render()


def test_kline_pro():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    bars = [RawBar(symbol=row['symbol'], open=row['open'], dt=row['dt'],
                   close=row['close'], high=row['high'], low=row['low'], vol=row['vol'])
            for _, row in kline.iterrows()]
    ka = CZSC(bars, freq="日线")

    # bs = []
    # for x in ka.bi_list:
    #     if x.fx_b.mark == Mark.D:
    #         mark = "buy"
    #     else:
    #         mark = "sell"
    #     bs.append({"dt": x.f, "mark": mark, mark: x.fx_b.fx})
    #
    # chart = plot.kline_pro(ka.bars_ubi, fx=ka.fx_list, bi=ka.bi_list, xd=ka.xd_list, bs=bs)
    # chart.render()




# coding: utf-8

import warnings
from typing import List
from .objects import Direction, BI
from .enum import Signals

def check_three_fd(fds: List[BI]) -> str:
    """识别三段形态
    :param fds: list
        由远及近的三段形态
    :return: str
    """
    v = Signals.Other.value

    if len(fds) != 3:
        return v

    fd1, fd2, fd3 = fds
    if not (fd1.direction == fd3.direction):
        print("1,3的 direction 不一致，无法识别三段形态")
        return v

    if fd3.direction == Direction.Down:
        if fd3.low > fd1.high:
            v = Signals.X3LA0.value

        if fd2.low < fd3.low < fd1.high < fd2.high:
            v = Signals.X3LB0.value

        if fd1.high > fd3.high and fd1.low < fd3.low:
            v = Signals.X3LC0.value

        if fd1.high < fd3.high and fd1.low > fd3.low:
            v = Signals.X3LD0.value

        if fd3.low < fd1.low and fd3.high < fd1.high:
            if fd3.power < fd1.power:
                v = Signals.X3LE0.value
            else:
                v = Signals.X3LF0.value
    elif fd3.direction == Direction.Up:
        if fd3.high > fd1.low:
            v = Signals.X3SA0.value

        if fd2.low < fd1.low < fd3.high < fd2.high:
            v = Signals.X3SB0.value

        if fd1.high > fd3.high and fd1.low < fd3.low:
            v = Signals.X3SC0.value

        if fd1.high < fd3.high and fd1.low > fd3.low:
            v = Signals.X3SD0.value

        if fd3.low > fd1.low and fd3.high > fd1.high:
            if fd3.power < fd1.power:
                v = Signals.X3SE0.value
            else:
                v = Signals.X3SF0.value
    else:
        raise ValueError("direction 的取值错误")

    return v

def check_five_fd(fds: List[BI]) -> str:
    """识别五段形态

    :param fds: 由远及近的五段走势
    :return: str
    """
    v = Signals.Other.value

    if len(fds) != 5:
        return v

    fd1, fd2, fd3, fd4, fd5 = fds
    if not (fd1.direction == fd3.direction == fd5.direction):
        print("1,3,5 的 direction 不一致，无法识别五段形态")
        return v
    direction = fd1.direction
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if direction == Direction.Down:
        if min(fd2.high, fd4.high) > max(fd2.low, fd4.low) and max_high == fd1.high and fd5.power < fd1.power:
            if (min_low == fd3.low and fd5.low < fd1.low) or (min_low == fd5.low):
                v = Signals.LA0.value

        if max(fd1.low, fd3.low) < min(fd1.high, fd3.high) < fd5.low:
            v = Signals.LI0.value

        if fd4.high < fd2.low and fd5.power < fd3.power and max_high == fd1.high:
            v = Signals.LF0.value

        if fd1.high < fd3.high < fd5.high and fd1.low > fd3.low > fd5.low:
            v = Signals.LJ0.value

        if fd1.high > fd3.high > fd5.high and fd1.low < fd3.low < fd5.low:
            v = Signals.LK0.value

        if (min_low == fd1.low and fd5.high > min(fd1.high, fd2.high) > fd5.low > fd1.low) \
                or (min_low == fd3.low and fd5.high > min(fd3.high, fd4.high) > fd5.low > fd3.low):
            v = Signals.LG0.value

    elif direction == Direction.Up:
        if min(fd2.high, fd4.high) > max(fd2.low, fd4.low) and min_low == fd1.low and fd5.power < fd1.power:
            if (max_high == fd3.high and fd5.high > fd1.high) or (max_high == fd5.high):
                v = Signals.SA0.value

        if min(fd1.high, fd3.high) > max(fd1.low, fd3.low) > fd5.high:
            v = Signals.SI0.value

        if min_low == fd1.low and fd5.power < fd1.power and fd4.low > fd2.high:
            v = Signals.SF0.value

        if fd1.high < fd3.high < fd5.high and fd1.low > fd3.low > fd5.low:
            v = Signals.SJ0.value

        if fd1.high > fd3.high > fd5.high and fd1.low < fd3.low < fd5.low:
            v = Signals.SK0.value

        if (max_high == fd1.high and fd5.low < max(fd1.low, fd2.low) < fd5.high < max_high) \
                or (max_high == fd3.high and fd5.low < max(fd3.low, fd4.low) < fd5.high < max_high):
            v = Signals.SG0.value

    else:
        raise ValueError("direction 的取值错误")
    return v

def check_seven_fd(fds: List[BI]) -> str:
    """识别七段形态

    :param fds: 由远及近的七段走势
    :return: str
    """
    v = Signals.Other.value
    if len(fds) != 7:
        return v

    fd1, fd2, fd3, fd4, fd5, fd6, fd7 = fds
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if fd7.direction == Direction.Down:
        if fd1.high == max_high and fd7.low == min_low:
            if min(fd2.high, fd4.high) > max(fd2.low, fd4.low) > fd6.high and fd7.power < fd5.power:
                v = Signals.LB0.value

            if fd2.low > min(fd4.high, fd6.high) > max(fd4.low, fd6.low) and fd7.power < (fd1.high - fd3.low):
                v = Signals.LD0.value

            if min(fd2.high, fd4.high, fd6.high) > max(fd2.low, fd4.low, fd6.low) and fd7.power < fd1.power:
                v = Signals.LA0.value

            if fd2.low > fd4.high and fd4.low > fd6.high and fd7.power < fd5.power:
                v = Signals.LF0.value

        if fd4.low == min_low and min(fd1.high, fd3.high) > max(fd1.low, fd3.low) \
                and min(fd5.high, fd7.high) > max(fd5.low, fd7.low) \
                and max(fd4.high, fd6.high) > min(fd3.high, fd4.high):
            if max(fd1.high, fd3.high) < max(fd5.high, fd7.high):
                v = Signals.LH1.value
            else:
                v = Signals.LH0.value

    elif fd7.direction == Direction.Up:
        if fd1.low == min_low and fd7.high == max_high:
            if fd6.low > min(fd2.high, fd4.high) > max(fd2.low, fd4.low) and fd7.power < fd5.power:
                v = Signals.SB0.value

            if min(fd4.high, fd6.high) > max(fd4.low, fd6.low) > fd2.high and fd7.power < (fd3.high - fd1.low):
                v = Signals.SD0.value

            if min(fd2.high, fd4.high, fd6.high) > max(fd2.low, fd4.low, fd6.low) and fd7.power < fd1.power:
                v = Signals.SA0.value

            if fd2.high < fd4.low and fd4.high < fd6.low and fd7.power < fd5.power:
                v = Signals.SF0.value

        if fd4.high == max_high and min(fd1.high, fd3.high) > max(fd1.low, fd3.low) \
                and min(fd5.high, fd7.high) > max(fd5.low, fd7.low) \
                and min(fd4.low, fd6.low) < max(fd3.low, fd4.low):
            if min(fd1.low, fd3.low) > min(fd5.low, fd7.low):
                v = Signals.SH1.value
            else:
                v = Signals.SH0.value

    else:
        raise ValueError("direction 的取值错误")
    return v

def check_nine_fd(fds: List[BI]) -> str:
    """识别九段形态

    :param fds: list
        由远及近的九段形态
    :return: str
    """
    v = Signals.Other.value
    if len(fds) != 9:
        return v

    direction = fds[-1].direction
    fd1, fd2, fd3, fd4, fd5, fd6, fd7, fd8, fd9 = fds
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if direction == Direction.Down:
        if min_low == fd9.low and max_high == fd1.high:
            # aAbBc式底背弛
            if min(fd2.high, fd4.high) > max(fd2.low, fd4.low) > fd6.high \
                    and min(fd6.high, fd8.high) > max(fd6.low, fd8.low) \
                    and min(fd2.low, fd4.low) > max(fd6.high, fd8.high) \
                    and fd9.power < fd5.power:
                v = Signals.LC0.value

            if min(fd2.high, fd4.high, fd6.high, fd8.high) > max(fd2.low, fd4.low, fd6.low, fd8.low) \
                    and fd9.power < fd1.power and fd3.low >= fd1.low and fd7.high <= fd9.high:
                v = Signals.LA0.value

            if min(fd2.high, fd4.high, fd6.high) > max(fd2.low, fd4.low, fd6.low) > fd8.high \
                    and fd9.power < fd7.power:
                v = Signals.LB0.value

            if fd3.low < fd1.low and fd7.high > fd9.high \
                    and min(fd4.high, fd6.high) > max(fd4.low, fd6.low) \
                    and (fd1.high - fd3.low) > (fd7.high - fd9.low):
                v = Signals.LE0.value

    elif direction == Direction.Up:
        if max_high == fd9.high and min_low == fd1.low:
            # aAbBc式顶背驰
            if fd6.low > min(fd2.high, fd4.high) > max(fd2.low, fd4.low) \
                    and min(fd6.high, fd8.high) > max(fd6.low, fd8.low) \
                    and max(fd2.high, fd4.high) < min(fd6.low, fd8.low) \
                    and fd9.power < fd5.power:
                v = Signals.SC0.value

            if min(fd2.high, fd4.high, fd6.high, fd8.high) > max(fd2.low, fd4.low, fd6.low, fd8.low) \
                    and fd9.power < fd1.power and fd3.high <= fd1.high and fd7.low >= fd9.low:
                v = Signals.SA0.value

            if fd8.low > min(fd2.high, fd4.high, fd6.high) > max(fd2.low, fd4.low, fd6.low) \
                    and fd9.power < fd7.power:
                v = Signals.SB0.value

            if fd3.high > fd1.high and fd7.low < fd9.low \
                    and min(fd4.high, fd6.high) > max(fd4.low, fd6.low) \
                    and (fd3.high - fd1.low) > (fd9.high - fd7.low):
                v = Signals.SE0.value

    else:
        raise ValueError("direction 的取值错误")
    return v

def check_eleven_fd(fds: List[BI]) -> str:
    """识别十一段形态

    :param fds: list
        由远及近的十一段形态
    :return: str
    """
    v = Signals.Other.value
    if len(fds) != 11:
        return v

    direction = fds[-1].direction
    fd1, fd2, fd3, fd4, fd5, fd6, fd7, fd8, fd9, fd10, fd11 = fds
    max_high = max([x.high for x in fds])
    min_low = min([x.low for x in fds])

    if direction == Direction.Down:
        if min_low == fd11.low and max_high == fd1.high:
            # aAbBc式底背弛，fd2-fd6构成A
            if min(fd2.high, fd4.high, fd6.high) > max(fd2.low, fd4.low, fd6.low) > fd8.high \
                    and min(fd8.high, fd10.high) > max(fd8.low, fd10.low) \
                    and min(fd2.low, fd4.low, fd6.low) > max(fd8.high, fd10.high) \
                    and fd11.power < fd7.power:
                # v = Signals.LC0.value
                v = Signals.LC1.value

            # aAbBc式底背弛，fd6-fd10构成B
            if min(fd2.high, fd4.high) > max(fd2.low, fd4.low) > fd6.high \
                    and min(fd6.high, fd8.high, fd10.high) > max(fd6.low, fd8.low, fd10.low) \
                    and min(fd2.low, fd4.low) > max(fd6.high, fd8.high, fd10.high) \
                    and fd11.power < fd5.power:
                # v = Signals.LC0.value
                v = Signals.LC1.value

            # ABC式底背驰，fd1-fd5构成A
            if fd5.low == min([x.low for x in [fd1, fd3, fd5]]) \
                    and fd9.low > fd11.low and fd9.high > fd11.high \
                    and fd8.high > fd6.low and fd1.high - fd5.low > fd9.high - fd11.low:
                # v = Signals.LE0.value
                v = Signals.LE1.value

            # ABC式底背驰，fd7-fd11构成C
            if fd1.high > fd3.high and fd1.low > fd3.low \
                    and fd7.high == max([x.high for x in [fd7, fd9, fd11]]) \
                    and fd6.high > fd4.low and fd1.high - fd3.low > fd7.high - fd11.low:
                # v = Signals.LE0.value
                v = Signals.LE1.value

    elif direction == Direction.Up:
        if max_high == fd11.high and min_low == fd1.low:
            # aAbBC式顶背驰，fd2-fd6构成A
            if fd8.low > min(fd2.high, fd4.high, fd6.high) >= max(fd2.low, fd4.low, fd6.low) \
                    and min(fd8.high, fd10.high) >= max(fd8.low, fd10.low) \
                    and max(fd2.high, fd4.high, fd6.high) < min(fd8.low, fd10.low) \
                    and fd11.power < fd7.power:
                # v = Signals.SC0.value
                v = Signals.SC1.value

            # aAbBC式顶背驰，fd6-fd10构成B
            if fd6.low > min(fd2.high, fd4.high) >= max(fd2.low, fd4.low) \
                    and min(fd6.high, fd8.high, fd10.high) >= max(fd6.low, fd8.low, fd10.low) \
                    and max(fd2.high, fd4.high) < min(fd6.low, fd8.low, fd10.low) \
                    and fd11.power < fd7.power:
                # v = Signals.SC0.value
                v = Signals.SC1.value

            # ABC式顶背驰，fd1-fd5构成A
            if fd5.high == max([fd1.high, fd3.high, fd5.high]) and fd9.low < fd11.low and fd9.high < fd11.high \
                    and fd8.low < fd6.high and fd11.high - fd9.low < fd5.high - fd1.low:
                # v = Signals.SE0.value
                v = Signals.SE1.value

            # ABC式顶背驰，fd7-fd11构成C
            if fd7.low == min([fd11.low, fd9.low, fd7.low]) and fd1.high < fd3.high and fd1.low < fd3.low \
                    and fd6.low < fd4.high and fd11.high - fd7.low < fd3.high - fd1.low:
                # v = Signals.SE0.value
                v = Signals.SE1.value
    else:
        raise ValueError("direction 的取值错误")
    return v




# coding: utf-8

from typing import List, Set

def has_interaction(v1: [List, Set], v2: [List, Set]) -> bool:
    """判断 v1 和 v2 是否存在交集

    :param v1:
    :param v2:
    :return:
    """
    if set(v1).intersection(set(v2)):
        return True
    else:
        return False




# coding: utf-8
import sys
import warnings

sys.path.insert(0, '.')
sys.path.insert(0, '..')
import os
import numpy as np
import pandas as pd
import czsc
from czsc.objects import RawBar
from czsc.cobra.utils import down_cross_count, kdj_gold_cross, kdj_dead_cross

warnings.warn("czsc version is {}".format(czsc.__version__))

cur_path = os.path.split(os.path.realpath(__file__))[0]


def test_kdj_cross():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    bars = [RawBar(symbol=row['symbol'], open=row['open'], dt=row['dt'],
                   close=row['close'], high=row['high'], low=row['low'], vol=row['vol'])
            for _, row in kline.iterrows()]

    assert not kdj_gold_cross(kline, just=False)
    assert not kdj_gold_cross(bars, just=False)
    assert kdj_dead_cross(kline, just=False)
    assert kdj_dead_cross(bars, just=False)
    assert not kdj_dead_cross(kline, just=True)


def test_cross_count():
    x1 = [1, 1, 3, 4, 5, 12, 9, 8]
    x2 = [2, 2, 1, 5, 8, 9, 10, 10]
    assert down_cross_count(x1, x2) == 2
    assert down_cross_count(np.array(x1), np.array(x2)) == 2
    assert down_cross_count(x2, x1) == 2
    assert down_cross_count(np.array(x2), np.array(x1)) == 2








# coding: utf-8
"""
常用技术分析指标：MA, MACD
"""
import numpy as np

def SMA(close: np.array, timeperiod=5):
    """简单移动平均

    https://baike.baidu.com/item/%E7%A7%BB%E5%8A%A8%E5%B9%B3%E5%9D%87%E7%BA%BF/217887

    :param close: np.array
        收盘价序列
    :param timeperiod: int
        均线参数
    :return: np.array
    """
    res = []
    for i in range(len(close)):
        if i < timeperiod:
            seq = close[0: i+1]
        else:
            seq = close[i - timeperiod + 1: i + 1]
        res.append(seq.mean())
    return np.array(res, dtype=np.double)

def EMA(close: np.array, timeperiod=5):
    """
    https://baike.baidu.com/item/EMA/12646151

    :param close: np.array
        收盘价序列
    :param timeperiod: int
        均线参数
    :return: np.array
    """
    res = []
    for i in range(len(close)):
        if i < 1:
            res.append(close[i])
        else:
            ema = (2 * close[i] + res[i-1] * (timeperiod-1)) / (timeperiod+1)
            res.append(ema)
    return np.array(res, dtype=np.double)

def MACD(close: np.array, fastperiod=12, slowperiod=26, signalperiod=9):
    """MACD 异同移动平均线
    https://baike.baidu.com/item/MACD%E6%8C%87%E6%A0%87/6271283

    :param close: np.array
        收盘价序列
    :param fastperiod: int
        快周期，默认值 12
    :param slowperiod: int
        慢周期，默认值 26
    :param signalperiod: int
        信号周期，默认值 9
    :return: (np.array, np.array, np.array)
        diff, dea, macd
    """
    ema12 = EMA(close, timeperiod=fastperiod)
    ema26 = EMA(close, timeperiod=slowperiod)
    diff = ema12 - ema26
    dea = EMA(diff, timeperiod=signalperiod)
    macd = (diff - dea) * 2
    return diff, dea, macd

def KDJ(close: np.array, high: np.array, low: np.array):
    """

    :param close: 收盘价序列
    :param high: 最高价序列
    :param low: 最低价序列
    :return:
    """
    n = 9
    hv = []
    lv = []
    for i in range(len(close)):
        if i < n:
            h_ = high[0: i+1]
            l_ = low[0: i+1]
        else:
            h_ = high[i - n + 1: i + 1]
            l_ = low[i - n + 1: i + 1]
        hv.append(max(h_))
        lv.append(min(l_))

    hv = np.around(hv, decimals=2)
    lv = np.around(lv, decimals=2)
    rsv = np.where(hv == lv, 0, (close - lv) / (hv - lv) * 100)

    k = []
    d = []
    j = []
    for i in range(len(rsv)):
        if i < n:
            k_ = rsv[i]
            d_ = k_
        else:
            k_ = (2 / 3) * k[i-1] + (1 / 3) * rsv[i]
            d_ = (2 / 3) * d[i-1] + (1 / 3) * k_

        k.append(k_)
        d.append(d_)
        j.append(3 * k_ - 2 * d_)

    k = np.array(k, dtype=np.double)
    d = np.array(d, dtype=np.double)
    j = np.array(j, dtype=np.double)
    return k, d, j

def RSQ(close: [np.array, list]) -> float:
    """拟合优度 R SQuare

    :param close: 收盘价序列
    :return:
    """
    x = list(range(len(close)))
    y = np.array(close)
    x_squred_sum = sum([x1 * x1 for x1 in x])
    xy_product_sum = sum([x[i] * y[i] for i in range(len(x))])
    num = len(x)
    x_sum = sum(x)
    y_sum = sum(y)
    delta = float(num * x_squred_sum - x_sum * x_sum)
    if delta == 0:
        return 0
    y_intercept = (1 / delta) * (x_squred_sum * y_sum - x_sum * xy_product_sum)
    slope = (1 / delta) * (num * xy_product_sum - x_sum * y_sum)

    y_mean = np.mean(y)
    ss_tot = sum([(y1 - y_mean) * (y1 - y_mean) for y1 in y]) + 0.00001
    ss_err = sum([(y[i] - slope * x[i] - y_intercept) * (y[i] - slope * x[i] - y_intercept) for i in range(len(x))])
    rsq = 1 - ss_err / ss_tot

    return round(rsq, 4)




# coding: utf-8
import os
from czsc.analyze import *
from czsc.utils.echarts_plot import kline_pro

cur_path = os.path.split(os.path.realpath(__file__))[0]
# cur_path = r"D:\git_repo\offline\dealer\test\czsc"

def test_find_bi():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    bars = [RawBar(symbol=row['symbol'], open=row['open'], dt=row['dt'],
                   close=row['close'], high=row['high'], low=row['low'], vol=row['vol'])
            for _, row in kline.iterrows()]

    # 去除包含关系
    bars1 = []
    for bar in bars:
        if len(bars1) < 2:
            bars1.append(NewBar(symbol=bar.symbol, dt=bar.dt, open=bar.open,
                                close=bar.close, high=bar.high, low=bar.low,
                                vol=bar.vol, elements=[bar]))
        else:
            k1, k2 = bars1[-2:]
            has_include, k3 = remove_include(k1, k2, bar)
            if has_include:
                bars1[-1] = k3
            else:
                bars1.append(k3)

    fxs = []
    for i in range(1, len(bars1)-1):
        fx = check_fx(bars1[i-1], bars1[i], bars1[i+1])
        if isinstance(fx, FX):
            fxs.append(fx)

def test_ka_update():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    bars = [RawBar(symbol=row['symbol'], open=row['open'], dt=row['dt'],
                   close=row['close'], high=row['high'], low=row['low'], vol=row['vol'])
            for _, row in kline.iterrows()]
    c = CZSC(bars, freq="日线", max_bi_count=20)

    kline = [x.__dict__ for x in c.bars_raw]
    bi = [{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in c.bi_list] + \
         [{'dt': c.bi_list[-1].fx_b.dt, "bi": c.bi_list[-1].fx_b.fx}]
    chart = kline_pro(kline, bi=bi, title="{} - {}".format(c.symbol, c.freq))
    chart.render("x.html")



# coding: utf-8
from setuptools import setup, find_packages
import czsc

setup(
    name="czsc",
    version=czsc.__version__,
    author=czsc.__author__,
    author_email=czsc.__email__,
    keywords=["缠论", "技术分析", "A股", "期货", "缠中说禅", "单因子分析"],
    description="缠中说禅技术分析工具",
    long_description="缠中说禅技术分析工具，源自 http://blog.sina.com.cn/chzhshch",
    license="MIT",

    url="https://github.com/zengbin93/czsc",
    packages=find_packages(exclude=['test', 'images', 'docs']),
    include_package_data=True,
    install_requires=["pandas", "pyecharts", "tushare", "requests", "seaborn"],

    classifiers=[
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
        ]
)



# coding: utf-8
import pandas as pd
from datetime import datetime, timedelta
from .factors import KlineGeneratorBy1Min, CzscFactors, factors_func
from ..data.jq import get_kline, get_kline_period
from ..data import freq_inv


class CzscTrader:
    """缠中说禅股票 选股/择时"""
    def __init__(self, symbol, max_count=1000, end_date=None):
        """
        :param symbol:
        """
        self.symbol = symbol
        if end_date:
            self.end_date = pd.to_datetime(end_date)
        else:
            self.end_date = datetime.now()
        self.max_count = max_count
        kg = KlineGeneratorBy1Min(max_count=max_count*2, freqs=['1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线'])
        for freq in kg.freqs:
            bars = get_kline(symbol, end_date=self.end_date, freq=freq_inv[freq], count=max_count)
            kg.init_kline(freq, bars)
        kf = CzscFactors(kg, factors=factors_func)
        self.kf = kf
        self.s = kf.s
        self.freqs = kg.freqs

    def __repr__(self):
        return "<CzscTrader of {} @ {}>".format(self.symbol, self.kf.end_dt)

    def take_snapshot(self, file_html, width="1400px", height="680px"):
        self.kf.take_snapshot(file_html, width, height)

    def open_in_browser(self, width="1400px", height="580px"):
        self.kf.open_in_browser(width, height)

    def update_factors(self):
        """更新K线数据到最新状态"""
        bars = get_kline_period(symbol=self.symbol, start_date=self.kf.end_dt, end_date=datetime.now(), freq="1min")
        if not bars:
            return
        for bar in bars:
            self.kf.update_factors(bar)
        self.s = self.kf.s

    def forward(self, n: int = 3):
        """向前推进N天"""
        ed = self.kf.end_dt + timedelta(days=n)
        if ed > datetime.now():
            print(f"{ed} > {datetime.now()}，无法继续推进")
            return

        bars = get_kline_period(symbol=self.symbol, start_date=self.kf.end_dt, end_date=ed, freq="1min")
        if not bars:
            print(f"{self.kf.end_dt} ~ {ed} 没有交易数据")
            return

        for bar in bars:
            self.kf.update_factors(bar)

        self.s = self.kf.s
        self.end_date = self.kf.end_dt.date()

